# encoding: UTF-8
#
# Copyright 2012-2013 Alejandro Autalán
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  http://pygubu.web.here

from __future__ import unicode_literals
import os
try:
    import tkinter as tk
    import tkinter.ttk as ttk
    import tkinter.filedialog
except:
    import Tkinter as tk
    import ttk
    import tkFileDialog
    tk.filedialog = tkFileDialog

from pygubu.stockimage import *
from pygubudesigner.widgets.compoundpropertyeditor import CompoundPropertyEditor
#from compoundpropertyeditor import CompoundPropertyEditor


class ImageEntry(CompoundPropertyEditor):

    def __init__(self, master=None, **kw):
        CompoundPropertyEditor.__init__(self, master, **kw)
        
        self._entryvar = tk.StringVar()
        self.entry = w = ttk.Entry(self, textvariable=self._entryvar)
        w.grid(sticky='ew')
        w.bind('<FocusOut>', self._on_entry_changed)
        w.bind('<KeyPress-Return>', self._on_entry_changed)
        w.bind('<KeyPress-KP_Enter>', self._on_entry_changed)
        self.button = w = ttk.Button(self, text='…',
            style='ImageSelectorButton.Toolbutton')
        w.grid(row=0, column=1, padx="5 0")
        w.configure(command=self._on_button_click)
        
        self.columnconfigure(0, weight=1)

    def _on_button_click(self):
        ext = ['*.gif', '*.pgm', '*.ppm']
        if tk.TkVersion >= 8.6:
            ext.append('*.png')
        options = {
            'filetypes':[('All Images', ' '.join(ext))] }

        fname = tk.filedialog.askopenfilename(**options)
        if fname:
            base, name = os.path.split(fname)
            if not StockImage.is_registered(name):
                StockImage.register(name, fname)
            self._disable_cb()
            self._variable.set(name)
            self._entryvar.set(name)
            self._enable_cb()
            self.event_generate('<<ImageEntryChanged>>')

    def _on_entry_changed(self, event=None):
        imagename = self._entryvar.get()
        self._disable_cb()
        self._variable.set(imagename)
        self._enable_cb()
        self.event_generate('<<ImageEntryChanged>>')

    def _on_variable_changed(self, varname, elementname, mode):
        imagename = self._variable.get()
        self._entryvar.set(imagename)

    
    
if __name__ == '__main__':
    root = tk.Tk()
    var = tk.StringVar()
    
    def see_var():
        print(var.get())

    entry = ImageEntry(root, textvariable=var)
    entry.grid()
    entry.configure(textvariable=var)
    var.set('red')
    
    btn = ttk.Button(root, text='Value', command=see_var)
    btn.grid(row=0, column=1)

    root.mainloop()




# encoding: UTF-8
#
# Copyright 2012-2013 Alejandro Autalán
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  http://pygubu.web.here

from __future__ import unicode_literals
try:
    import tkinter as tk
    import tkinter.ttk as ttk
except:
    import Tkinter as tk
    import ttk
    
    
class CompoundPropertyEditor(ttk.Frame):
    def __init__(self, master=None, **kw):
        self._variable = None
        self.__cbhandler = None
        
        if 'textvariable' in kw:
            self._configure_variable(kw.pop('textvariable'))
        ttk.Frame.__init__(self, master, **kw)
        
    def configure(self, cnf=None, **kw):
        if 'textvariable' in kw:
            self._configure_variable(kw.pop('textvariable'))
        ttk.Frame.configure(self, cnf, **kw)

    config = configure
        
    def _configure_variable(self, variable):
        self._disable_cb()
        self._variable = variable
        self._enable_cb()

    def _enable_cb(self):
        if self._variable is not None:
            self.__cbhandler = self._variable.trace(
                mode="w", callback=self._on_variable_changed)

    def _disable_cb(self):
        if self.__cbhandler is not None:
            self._variable.trace_vdelete("w", self.__cbhandler)
            self.__cbhandler = None
            
    def _on_variable_changed(self, varname, elementname, mode):
        pass



# encoding: UTF-8
#
# Copyright 2012-2013 Alejandro Autalán
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  http://pygubu.web.here

from __future__ import unicode_literals
try:
    import tkinter as tk
    import tkinter.ttk as ttk
except:
    import Tkinter as tk
    import ttk

from pygubudesigner.widgets.compoundpropertyeditor import CompoundPropertyEditor

class Textentry(CompoundPropertyEditor):
    def __init__(self, master=None, cnf={}, **kw):
        CompoundPropertyEditor.__init__(self, master, **kw)
        self.text = w = tk.Text(self)
        w.grid(sticky='nsew')
        w.bind('<FocusOut>', self._on_focus_out)
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

    def _on_focus_out(self, event=None):
        value = (self.text.get('0.0', tk.END))[:-1]
        if self._variable is not None:
            self._disable_cb()
            self._variable.set(value)
            self._enable_cb()
            self.event_generate('<<TextentryChanged>>')

    def _on_variable_changed(self, varname, elementname, mode):
        self.text.delete('0.0', tk.END)
        self.text.insert('0.0', self._variable.get())




from __future__ import unicode_literals
try:
    import tkinter as tk
    import tkinter.ttk as ttk
except:
    import Tkinter as tk
    import ttk

class BindingsEditor:
    def __init__(self, etreeview):
        self.tv = etreeview
        self._curr_data = None
        self._adder = 'adder'
        self._allow_edit = False
        self.tv.insert('', tk.END, iid=self._adder, values=('+',))
        self.tv.bind('<<TreeviewInplaceEdit>>', self._on_inplace_edit)
        self.tv.bind('<<TreeviewCellEdited>>', self._on_cell_edited)
        self.tv.bind('<Double-Button-1>', self._on_add_clicked, add=True)

        self._del_btn = ttk.Button(self.tv, text='-',
            command=self._on_del_clicked)

    def _on_add_clicked(self, event):
        sel =  self.tv.selection()
        if sel:
            item = sel[0]
            if item == self._adder and self._allow_edit:
                self._add_binding(('<1>', 'callback', ''))

    def _on_del_clicked(self):
        sel =  self.tv.selection()
        if sel:
            item = sel[0]
            self.tv.delete(item)
        self._copy_to_data()

    def _on_cell_edited(self, event):
        col, item = self.tv.get_event_info()
        self._copy_to_data()

    def _copy_to_data(self):
        if self._curr_data is not None:
            items = self.tv.get_children()
            self._curr_data.clear_bindings()
            for item in items:
                if item != self._adder:
                    values = self.tv.item(item, 'values')
#                    print(item, 'values = ', values)
                    self._curr_data.add_binding(*values[:3])
                    self._curr_data.notify(self)

    def _add_binding(self, bind):
        idx = self.tv.index(self._adder)
        item = self.tv.insert('', idx, values=bind)
        self.tv.selection_set('')

    def edit(self, data):
        self.clear()
        self._curr_data = data
        for bind in data.get_bindings():
            self._add_binding(bind)

    def allow_edit(self, var):
        self._allow_edit = var


    def _on_inplace_edit(self, event):
        col, item = self.tv.get_event_info()
#        print('on_inplace_edit:', col, item)
        if item != self._adder and self._allow_edit:
            if col in ('sequence', 'handler'):
                self.tv.inplace_entry(col, item)
            elif col in ('add',):
                self.tv.inplace_checkbutton(col, item, offvalue='')
            elif col in ('actions',):
                self.tv.inplace_custom(col, item, self._del_btn)


    def clear(self):
        children = self.tv.get_children()
        for item in children:
            if item != self._adder:
                self.tv.delete(item)



# encoding: UTF-8
#
# Copyright 2012-2013 Alejandro Autalán
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  http://pygubu.web.here

from __future__ import unicode_literals
try:
    import tkinter as tk
    import tkinter.ttk as ttk
except:
    import Tkinter as tk
    import ttk

from pygubudesigner.widgets.compoundpropertyeditor import CompoundPropertyEditor
#from compoundpropertyeditor import CompoundPropertyEditor


class TkVarEntry(CompoundPropertyEditor):
    
    def __init__(self, master=None, **kw):
        CompoundPropertyEditor.__init__(self, master, **kw)
        
        self._entryvar = tk.StringVar()
        self._cboxvar = var2 = tk.StringVar()
        var2.trace(mode="w", callback=self.__on_cbox_changed)
        self.entry = entry = ttk.Entry(self, textvariable=self._entryvar)
        entry.bind('<FocusOut>', self.__on_entry_changed)
        entry.bind('<KeyPress-Return>', self.__on_entry_changed)
        entry.bind('<KeyPress-KP_Enter>', self.__on_entry_changed)
        self.entry.grid(sticky='ew')
        
        self.default_vtype = 'string'
        cbox_values = ('string', 'int', 'double', 'boolean')
        self._cboxvar.set(cbox_values[0])
        self.cbox = ttk.Combobox(self, textvariable=self._cboxvar,
            state='readonly', values=cbox_values, width=6)
        self.cbox.grid(row=0, column=1)
        
        self.columnconfigure(0, weight=1)
        
    def __on_cbox_changed(self, varname, elementname, mode):
        self._on_value_changed_by_user()

    def __on_entry_changed(self, event=None):
        self._on_value_changed_by_user()

    def _on_value_changed_by_user(self):
        value = self._entryvar.get()
        if value:
            value = self._cboxvar.get() + ':' + value
        else:
            value = ''
        self._disable_cb()
        self._variable.set(value)
        self._enable_cb()
        self.event_generate('<<TkVarEntryChanged>>')
        
    def _on_variable_changed(self, varname, elementname, mode):
        vtype = self.default_vtype
        value = self._variable.get()
        if ':' in value:
            vtype, value = value.split(':')
        self._cboxvar.set(vtype)
        self._entryvar.set(value)

    
    
if __name__ == '__main__':
    root = tk.Tk()
    var = tk.StringVar()
    
    def see_var():
        print(var.get())

    entry = TkVarEntry(root, textvariable=var)
    entry.grid()
    entry.configure(textvariable=var)
    var.set('Double:mydouble')
    
    btn = ttk.Button(root, text='Value', command=see_var)
    btn.grid(row=0, column=1)

    root.mainloop()

    



# encoding: UTF-8
#
# Copyright 2012-2013 Alejandro Autalán
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  http://pygubu.web.here
from __future__ import unicode_literals
import re
from collections import defaultdict
try:
    import tkinter as tk
    import tkinter.ttk as ttk
except:
    import Tkinter as tk
    import ttk

from pygubu import builder
from . import util
from . import properties
from .widgets import ColorEntry, Textentry, TkVarEntry, ImageEntry
from .widgets import SizeEntry
from .bindingseditor import BindingsEditor
from .i18n import translator as _


CLASS_MAP = builder.CLASS_MAP

FLOAT_RE = re.compile(r'[+-]?(\d+(\.\d*)?)')


class PropertiesArray(util.ArrayVar):

    def get_property_variable(self, group, name):
        if group not in properties.GROUPS:
            raise ValueError('Invalid group value')
        return self.__call__(group + name)

    def identify_property(self, element):
        result = (None, None)
        for groupname in properties.GROUPS:
            if element.startswith(groupname):
                result = (groupname, element.replace(groupname, ''))
                break
        return result


class WidgetPropertiesEditor:

    def __init__(self, app):
        self.current = None #data to edit
        #self.app = app
        #self.treeview = app.treeview
        self.properties_sf = app.builder.get_object('properties_sf')
        self.propsframe = app.widget_props_frame
        self.layoutframe_sf = app.builder.get_object('layoutframe_sf')
        self.layoutframe = app.layout_props_frame
        self.bindingsframe = app.bindings_frame
        self.arrayvar = PropertiesArray()
        self.prop_widget = {
            properties.GROUP_WIDGET: {},
            properties.GROUP_LAYOUT_GRID: {},
            properties.GROUP_LAYOUT_GRID_RC: {},
            properties.GROUP_CUSTOM: {},
            'internal': {}
        }
        self._editor_ready = False
        self._var_prev_value = {}
        #register validators
        tkwidget = self.propsframe
        self.validators = {
            'number_integer': (tkwidget.register(self.validator_integer),
                '%d', '%P'),
            'number_float': (tkwidget.register(self.validator_float),
                '%d', '%P'),
            'alphanumeric': (tkwidget.register(self.validator_alphanumeric),
                '%d', '%P'),
            'tkpadding': (tkwidget.register(self.validator_tkpadding4),
                '%d', '%P'),
            'tkpadding2': (tkwidget.register(self.validator_tkpadding2),
                '%d', '%P'),
            'entry_validate_args': (tkwidget.register(
                self.validator_entry_validate_args), '%d', '%P'),
        }
        self.create_properties()
        self.create_grid_layout_editor()

        self.bindings_editor = BindingsEditor(app.bindings_tree)

        self.hide_all()


    def validator_tkpadding4(self, action, newvalue):
        return self.validator_tkpadding(action, newvalue, 4)

    def validator_tkpadding2(self, action, newvalue):
        return self.validator_tkpadding(action, newvalue, 2)


    def validator_tkpadding(self, action, newvalue, maxitems):
        valid = True
        if action == '1': #1: insert 0: delete
            nums = newvalue.split()
            if len(nums) <= maxitems:
                for num in nums:
                    if not num.isdigit():
                        valid = False
                        break;
            else:
                valid = False
        else:
            valid = True
        return valid


    def validator_alphanumeric(self, action, newvalue):
        valid = False
        if action == '1': #1: insert 0: delete
            valid = str(newvalue).isalnum()
        else:
            valid = True
        return valid


    def validator_integer(self, action, newvalue):
        valid = False
        if action == '1': #1: insert 0: delete
            valid = str(newvalue).isalnum()
        else:
            valid = True
        return valid


    def validator_float(self, action, newvalue):
        valid = False
        if action == '1': #1: insert 0: delete
            match = FLOAT_RE.match(newvalue)
            if match is not None:
                valid = True
        else:
            valid = True
        return valid


    def validator_entry_validate_args(self, action, newvalue):
        valid = True
        valid_args = ('%', '%d', '%i', '%P', '%s', '%S', '%v', '%V', '%W')
        if action == '1': #1: insert 0: delete
            args = newvalue.split()
            for arg in args:
                if arg not in valid_args:
                    valid = False
                    break
        else:
            valid = True
        return valid


    def connect_focusout_cb(self, widget, pname):
        def focusout_handler(event, self=self, pname=pname):
            self.on_property_changed(pname)
        widget.bind('<FocusOut>', focusout_handler, add='+')


    def connect_on_enterkey_cb(self, widget, pname):
        def enterkey_handler(event, self=self, pname=pname):
            self.on_property_changed(pname)
        widget.bind('<KeyPress-Return>', enterkey_handler, add='+')
        widget.bind('<KeyPress-KP_Enter>', enterkey_handler, add='+')


    def connect_command_cb(self, widget, pname):
        def command_handler(self=self, pname=pname):
            self.on_property_changed(pname)
        widget.configure(command=command_handler)


    def connect_virtualevent_cb(self, eventname, widget, pname):
        def virtualevent_handler(event, self=self, pname=pname):
            self.on_property_changed(pname)
        widget.bind(eventname, virtualevent_handler, add='+')


    def connect_variable_cb(self, variable, pname):
        def var_handler(varname, elementname, mode, self=self, pname=pname):
            self.on_property_changed(pname)
        variable.trace('w', callback=var_handler)


    def on_property_changed(self, pname):
        '''Updates treeview values from property editor.'''

        #Do not redraw if editor is not ready. See edit method.
        if not self._editor_ready:
            return

        new_value = self.arrayvar[pname]
        old_value = self._var_prev_value.get(pname, '')

        #Do not redraw if same values
        #print('new ', new_value, ' old ', old_value)
        if new_value == old_value:
            return
        self._var_prev_value[pname] = new_value

        pgroup, pname = self.arrayvar.identify_property(pname)
        if self.current is not None:
            data = self.current
            widget_id = data.get_id()
            wclass = data.get_class()

            if (pgroup in (properties.GROUP_WIDGET, properties.GROUP_CUSTOM)):

                if new_value:
                    data.set_property(pname, new_value)
                else:
                    data.set_property(pname, '')

            elif (pgroup == properties.GROUP_LAYOUT_GRID and pname in
                    properties.PropertiesMap[properties.GROUP_LAYOUT_GRID]):

                if new_value:
                    data.set_layout_property(pname, new_value)
                else:
                    data.set_layout_property(pname, '')

            elif (pgroup == properties.GROUP_LAYOUT_GRID_RC):
                row_col, number, name = self.identify_gridrc_property(pname)
                number = str(number)
                if row_col == 'row':
                    if new_value:
                        data.set_grid_row_property(number, name, new_value)
                    else:
                        data.set_grid_row_property(number, name, '')
                elif row_col == 'column':
                    if new_value:
                        data.set_grid_col_property(number, name, new_value)
                    else:
                        data.set_grid_col_property(number, name, '')

            data.notify(self)


    def create_properties(self):
        """Populate a frame with a list of all editable properties"""

        editor_frame = self.propsframe
        prop_widget = self.prop_widget

        #hack to resize correctly when properties are hidden
        label = ttk.Label(editor_frame)
        label.grid()

        row=0
        col=0
        label_tpl = "{0}:"
        group = properties.GROUP_CUSTOM
        for name in properties.PropertiesMap[group]:
            labeltext = label_tpl.format(name)
            label = ttk.Label(editor_frame, text=labeltext, anchor=tk.W)
            togrid, widget = self.create_property_widget(editor_frame,
                group, name)
            label.grid(row=row, column=col, sticky=tk.EW, pady=2)
            togrid.grid(row=row, column=col+1, sticky=tk.EW, pady=2)
            row += 1
            prop_widget[group][name] = (label, togrid)

        group = properties.GROUP_WIDGET
        for name in properties.PropertiesMap[group]:
            labeltext = label_tpl.format(name)
            label = ttk.Label(editor_frame, text=labeltext, anchor=tk.W)
            togrid, widget = self.create_property_widget(editor_frame,
                group, name)
            label.grid(row=row, column=col, sticky=tk.EW, pady=2)
            togrid.grid(row=row, column=col+1, sticky=tk.EW, pady=2)
            row += 1
            prop_widget[group][name] = (label, togrid)


    def create_grid_layout_editor(self):
        master = self.layoutframe
        prop_widget = self.prop_widget
        frame = ttk.LabelFrame(master, text=_('Grid options:'), padding=5)

        #hack to resize correctly when properties are hidden
        label = ttk.Label(frame)
        label.grid()

        label_tpl = "{0}:"
        row= col=0
        group = properties.GROUP_LAYOUT_GRID
        for name in properties.PropertiesMap[group]:
            labeltext = label_tpl.format(name)
            label = ttk.Label(frame, text=labeltext, anchor=tk.W)
            togrid, widget = self.create_property_widget(frame, group, name)
            label.grid(row=row, column=col, sticky=tk.EW, pady=2)
            togrid.grid(row=row, column=col+1, sticky=tk.EW, pady=2)
            row += 1
            prop_widget[group][name] = (label, togrid)

        frame.grid(row=0, column=0, sticky=tk.NSEW)

        #labels
        group = properties.GROUP_LAYOUT_GRID_RC
        frame = ttk.LabelFrame(master, text=_('Grid row/column options:'),
            padding=5)

        #hack to resize correctly when properties are hidden
        label = ttk.Label(frame)
        label.grid()

        row = col = 0
        icol = 1
        headers=[]
        for pname in properties.PropertiesMap[group]:
            label = ttk.Label(frame, text=pname)
            label.grid(row=row, column=icol)
            headers.append(label)
            icol += 1
        prop_widget['internal']['grid_rc_headers']= headers

        #
        name_format = '{}_{}_{}' #{row/column}_{number}_{name}
        max_rc = 50
        #rowconfig
        row += 1
        trow_label = _('Row {0}:')
        tcol_label = _('Column {0}:')
        for index in range(0, max_rc):
            labeltext = trow_label.format(index)
            label = ttk.Label(frame, text=labeltext)
            label.grid(row=row, column=0)

            labeltext = tcol_label.format(index)
            labelc = ttk.Label(frame, text=labeltext)
            labelc.grid(row=row+max_rc, column=0, sticky=tk.E, pady=2)

            icol = 1
            for pname in properties.PropertiesMap[group]:
                alias = name_format.format('row', index, pname)
                togrid, widget = self.create_grid_rc_widget(frame, pname, alias)
                togrid.grid(row=row, column=icol, pady=2)
                prop_widget[group][alias]= (label, togrid)

                alias = name_format.format('column', index, pname)
                togrid, widget = self.create_grid_rc_widget(frame, pname, alias)
                togrid.grid(row=row+max_rc, column=icol, pady=2)
                prop_widget[group][alias]= (labelc, togrid)

                icol += 1
            row +=1

        frame.grid(row=1, column=0, sticky=tk.NSEW, pady='10 0')


    def create_grid_rc_widget(self, master, name, alias):
        group = properties.GROUP_LAYOUT_GRID_RC
        widgetvar = self.arrayvar.get_property_variable(group, alias);
        wdata = {}
        widget = None
        if name in properties.PropertiesMap[group]:
            wdata = properties.PropertiesMap[group][name]
            wtype = wdata['input_method']

        togrid, widget = self.__create_widget(master, wdata, group+alias,
            widgetvar)
        widget.configure(width=4)
        return (togrid, widget)


    def create_property_widget(self, master, group, propertyname):
        """Creates a ui widget to edit the property"""

        widgetvar = self.arrayvar.get_property_variable(group, propertyname);
        wdata = {}

        widget = None
        if propertyname in properties.PropertiesMap[group]:
            wdata = properties.PropertiesMap[group][propertyname]

        togrid, widget = self.__create_widget(master, wdata,
            group+propertyname, widgetvar)
        return (togrid, widget)


    def __create_widget(self, master, wdata, propalias, widgetvar):
        widget = None
        togrid = None
        wtype = wdata.get('input_method', None)
        default = wdata.get('default', '')

        #I don't have class name at this moment
        #so setup class specific values on update_property_widget
        if wtype == 'entry':
            readonly = wdata.get('readonly', False)
            state = tk.DISABLED if readonly else tk.NORMAL
            widget = ttk.Entry(master, textvariable=widgetvar, state=state)
            togrid = widget
            #validator
            validator = wdata.get('validator', None)
            if validator is not None:
                widget.configure(validate='key',
                    validatecommand=self.validators[validator])
            self.connect_focusout_cb(widget, propalias)
            self.connect_on_enterkey_cb(widget, propalias)
        elif wtype == 'textentry':
            widget = Textentry(master, textvariable=widgetvar)
            widget.text.configure(width=20, height=3)
            togrid = widget
            self.connect_virtualevent_cb('<<TextentryChanged>>', widget, propalias)
        elif wtype == 'colorentry':
            widget = ColorEntry(master, textvariable=widgetvar)
            togrid = widget
            self.connect_virtualevent_cb('<<ColorEntryChanged>>', widget, propalias)
        elif wtype == 'imageentry':
            widget= ImageEntry(master, textvariable=widgetvar)
            togrid = widget
            self.connect_virtualevent_cb('<<ImageEntryChanged>>', widget, propalias)
        elif wtype == 'choice':
            readonly = wdata.get('readonly', False)
            status = 'readonly' if readonly else tk.NORMAL
            widget = ttk.Combobox(master, textvariable=widgetvar,
                state=status)
            togrid = widget
            values = wdata.get('values', None)
            if values is not None:
                widget.configure(values=values)
            self.connect_variable_cb(widgetvar, propalias)
        elif wtype == 'tkvarentry':
            widget = TkVarEntry(master, textvariable=widgetvar,
                        width=20, height=3)
            togrid = widget
            self.connect_virtualevent_cb('<<TkVarEntryChanged>>', widget, propalias)
        elif wtype == 'spinbox':
            status= tk.NORMAL
            readonly = wdata.get('readonly', False)
            if readonly:
                status = 'readonly'
            widget = tk.Spinbox(master, textvariable=widgetvar,
                        state=status, readonlybackground='white')
            togrid = widget
            vmin = wdata.get('min', 0)
            vmax = wdata.get('max', 99)
            increment = wdata.get('increment', 1)
            default = vmin if (vmin > 0 and default == '') else default
            widget.configure(from_=vmin, to=vmax, increment=increment)
            #validator
            validator = wdata.get('validator', None)
            if validator is not None:
                widget.configure(validate='key',
                    validatecommand=self.validators[validator])
            self.connect_focusout_cb(widget, propalias)
            self.connect_command_cb(widget, propalias)
            self.connect_on_enterkey_cb(widget, propalias)
        elif wtype == 'sizeentry':
            widget = togrid = SizeEntry(master, textvariable=widgetvar)
            self.connect_virtualevent_cb('<<SizeEntryChanged>>',
                widget, propalias)
        else:
            widget = ttk.Entry(master, textvariable=widgetvar)
            togrid = widget
            self.connect_focusout_cb(widget, propalias)
            self.connect_on_enterkey_cb(widget, propalias)
        widgetvar.set(default)

        return (togrid, widget)


    def hide_all(self):
        """Hide all properties from property editor."""
        self.current = None
        for group in self.prop_widget:
            for pname in self.prop_widget[group]:
                for widget in self.prop_widget[group][pname]:
                    widget.grid_remove()
        self.properties_sf.reposition()
        self.layoutframe_sf.reposition()

        self.bindingsframe.grid_remove()
        self.bindings_editor.clear()


    def update_property_widget(self, group, widget, variable,
        propertyname, classname, data, roc=None, rc_id=None):
        """Update widget property value with values from data."""

        wdata = {}
        if propertyname in properties.PropertiesMap[group]:
            wdata = properties.PropertiesMap[group][propertyname]

        wtype = wdata.get('input_method', '')

        #merge description for specific class
        if classname in wdata:
            wdata = dict(wdata, **wdata[classname])

        default = wdata.get('default', '')
        if wtype == 'entry':
            #switch validator mode for entry
            validator = wdata.get('validator', None)
            if validator is not None:
                widget.configure(validate='key',
                    validatecommand=self.validators[validator])
        if wtype == 'spinbox':
            vmin = wdata.get('min', 0)
            vmax = wdata.get('max', 99)
            default = vmin if (int(vmin) > 0 and default == '') else default
        if wtype == 'choice':
            readonly = wdata.get('readonly', False)
            status = 'readonly' if readonly else tk.NORMAL
            values = wdata.get('values', None)
            if values is not None:
                widget.configure(values=values, state=status)
        if wtype == 'sizeentry':
            mode = wdata.get('mode', '')
            widget.set_mode(mode)

        value = ''
        if group in (properties.GROUP_CUSTOM, properties.GROUP_WIDGET):
            value = data.get_property(propertyname)
        elif group == properties.GROUP_LAYOUT_GRID:
            value = data.get_layout_property(propertyname)
        elif group == properties.GROUP_LAYOUT_GRID_RC:
            if roc == 'row':
                value = data.get_grid_row_property(rc_id, propertyname)
            else:
                value = data.get_grid_col_property(rc_id, propertyname)

        if not value and default:
            value = default

        variable.set(value)
        self._var_prev_value[group+propertyname] = value


    def edit(self, data, max_row=1, max_col=1):
        """Copies properties values from data to the
           properties editor so they can be edited."""

        self._var_prev_value.clear()
        #
        self._editor_ready = False
        self.current = data

        wclass = data.get_class()
        class_props = tuple(CLASS_MAP[wclass].classobj.properties)

        group = properties.GROUP_CUSTOM
        for key in properties.PropertiesMap[group]:
            if (key in properties.OBJECT_DEFAULT_ATTRS or key in class_props):
                label, widget = self.prop_widget[group][key]
                variable = self.arrayvar.get_property_variable(group, key)
                self.update_property_widget(group, widget, variable,
                    key, wclass, data)
                label.grid()
                widget.grid()
            else:
                #hide property widget
                label, widget = self.prop_widget[group][key]
                label.grid_remove()
                widget.grid_remove()

        group = properties.GROUP_WIDGET
        for key in properties.PropertiesMap[group]:
            if key in class_props:
                label, widget = self.prop_widget[group][key]
                variable = self.arrayvar.get_property_variable(group, key)
                self.update_property_widget(group, widget, variable,
                    key, wclass, data)
                label.grid()
                widget.grid()
            else:
                #hide property widget
                label, widget = self.prop_widget[group][key]
                label.grid_remove()
                widget.grid_remove()

        max_children = CLASS_MAP[wclass].classobj.maxchildren
        max_children = 0 if max_children is None else max_children
        is_container = CLASS_MAP[wclass].classobj.container
        layout_required = CLASS_MAP[wclass].classobj.layout_required
        allow_container_layout = CLASS_MAP[wclass].classobj.allow_container_layout
        show_layout = layout_required
        #grid layout properties
        group = properties.GROUP_LAYOUT_GRID
        for gkey in properties.PropertiesMap[group]:
            label, widget = self.prop_widget[group][gkey]
            if show_layout:
                variable = self.arrayvar.get_property_variable(group, gkey)
                self.update_property_widget(group, widget, variable, gkey,
                        wclass, data)
                label.grid()
                widget.grid()
            else:
                label.grid_remove()
                widget.grid_remove()

        group = properties.GROUP_LAYOUT_GRID_RC
        #max_row, max_col = self.get_max_row_col(item)
        show_layout = (is_container
                        and layout_required
                        and allow_container_layout
                        and max_children != 1)
        show_headers = False
        for key in self.prop_widget[group]:
            rowcol, number, name = self.identify_gridrc_property(key)
            number_str = str(number)
            number = int(number)
            if show_layout and rowcol == 'row' and number <= max_row:
                label, widget = self.prop_widget[group][key]
                variable = self.arrayvar.get_property_variable(group, key)
                self.update_property_widget(group, widget, variable, name,
                    wclass, data, rowcol, number_str)
                label.grid(); widget.grid()
                show_headers = True
            elif show_layout and rowcol == 'column' and number <= max_col:
                label, widget = self.prop_widget[group][key]
                variable = self.arrayvar.get_property_variable(group, key)
                self.update_property_widget(group, widget, variable, name,
                    wclass, data, rowcol, number_str)
                label.grid(); widget.grid()
                show_headers = True
            else:
                label, widget = self.prop_widget[group][key]
                label.grid_remove(); widget.grid_remove()

        for widget in self.prop_widget['internal']['grid_rc_headers']:
            if show_headers:
                widget.grid()
            else:
                widget.grid_remove()
        #
        self.properties_sf.reposition()
        self.layoutframe_sf.reposition()
        #
        allow_edit = CLASS_MAP[wclass].classobj.allow_bindings
        if allow_edit:
            self.bindingsframe.grid()
        else:
            self.bindingsframe.grid_remove()
        self.bindings_editor.allow_edit(allow_edit)
        self.bindings_editor.edit(data)

        self._editor_ready = True


    def identify_gridrc_property(self, alias):
        return alias.split('_')



# encoding: UTF-8
#
# Copyright 2012-2013 Alejandro Autalán
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  http://pygubu.web.here
from __future__ import unicode_literals
from collections import OrderedDict
try:
    import tkinter as tk
    import tkinter.ttk as ttk
except:
    import Tkinter as tk
    import ttk



_default_entry_prop = {
    'input_method': 'entry',
}

_default_textentry_prop = {
    'input_method': 'textentry',
}

_default_color_prop = {
    'input_method': 'colorentry',
}

_default_image_prop = {
    'input_method': 'imageentry',
}

_default_spinbox_prop = {
    'input_method': 'spinbox',
    'min': 0,
    'max': 999,
    'validator': 'number_integer',
}

_default_spinbox_float_prop = {
    'input_method': 'spinbox',
    'min': 0,
    'max': 999,
    'validator': 'number_float',
    'increment': 0.1
}

_default_tkvariable_prop = {
    'input_method': 'tkvarentry',
}

_default_dimension_prop = {
    'input_method': 'sizeentry',
}

_default_relief_prop = {
    'input_method': 'choice',
    'values': ('', tk.FLAT, tk.RAISED, tk.SUNKEN,
        tk.GROOVE, tk.RIDGE), 'readonly': True
}

_empty_choice = {'input_method': 'choice', 'readonly': True }

_default_true_false_prop = { 'input_method': 'choice',
    'values': ('', tk.TRUE, tk.FALSE), 'readonly': True}

_default_cursor_prop = {
    'input_method': 'choice',
    'values': ('', 'arrow', 'based_arrow_down','based_arrow_up','boat',
       'bogosity','bottom_left_corner','bottom_right_corner',
        'bottom_side', 'bottom_tee', 'box_spiral', 'center_ptr', 'circle',
        'clock', 'coffee_mug', 'cross', 'cross_reverse', 'crosshair',
        'diamond_cross', 'dot', 'dotbox', 'double_arrow',  'draft_large',
        'draft_small', 'draped_box', 'exchange', 'fleur', 'gobbler',
        'gumby', 'hand1', 'hand2', 'heart', 'icon', 'iron_cross',
        'left_ptr', 'left_side', 'left_tee', 'leftbutton', 'll_angle',
        'lr_angle', 'man', 'middlebutton', 'mouse', 'pencil', 'pirate',
        'plus', 'question_arrow', 'right_ptr', 'right_side', 'right_tee',
        'rightbutton', 'rtl_logo', 'sailboat', 'sb_down_arrow',
        'sb_h_double_arrow', 'sb_left_arrow', 'sb_right_arrow',
        'sb_up_arrow', 'sb_v_double_arrow', 'shuttle', 'sizing', 'spider',
        'spraycan', 'star', 'target', 'tcross', 'top_left_arrow',
        'top_left_corner', 'top_right_corner', 'top_side', 'top_tee',
        'trek', 'ul_angle', 'umbrella', 'ur_angle', 'watch', 'xterm',
        'X_cursor'),
    'readonly': True
}

_sticky_prop = {
        'input_method': 'choice',
        'values': ('', tk.N, tk.S,
            tk.E, tk.W,
            tk.NE, tk.NW,
            tk.SE, tk.SW,
            tk.EW, tk.NS,
            tk.NS + tk.W,
            tk.NS + tk.E,
            #tk.NSEW
            tk.NE + tk.SW
            ),
        'tk.Frame': {'default': tk.NE + tk.SW},
        'ttk.Frame': {'default': tk.NE + tk.SW},
        'tk.LabelFrame': {'default': tk.NE + tk.SW},
        'ttk.Labelframe': {'default': tk.NE + tk.SW},
        'tk.ScrollbarHelper': {'default': tk.NE + tk.SW},
        'ttk.ScrollbarHelper': {'default': tk.NE + tk.SW},
        'tk.ScrolledFrame': {'default': tk.NE + tk.SW},
        'ttk.ScrolledFrame': {'default': tk.NE + tk.SW},
        'ttk.Notebook': {'default': tk.NE + tk.SW}
        }

GROUP_WIDGET = 'widget__'
GROUP_LAYOUT_GRID = 'layoutgrid__'
GROUP_LAYOUT_GRID_RC = 'layoutgridrc__'
GROUP_CUSTOM = 'custom__'

GROUPS = (GROUP_WIDGET, GROUP_LAYOUT_GRID, GROUP_LAYOUT_GRID_RC, GROUP_CUSTOM)


PropertiesMap = {}

__widget = (
    ('accelerator', _default_entry_prop),
    ('activerelief', _default_relief_prop),
    ('activestyle', {
        'input_method': 'choice',
        'values': ('', 'underline', 'dotbox', 'none')
        }),
    ('activebackground', _default_color_prop),
    ('activeborderwidth', _default_spinbox_prop),
    ('activeforeground', _default_color_prop),
    ('after', _empty_choice),
    ('anchor', {
        'input_method': 'choice',
        'values': ('', tk.W, tk.CENTER, tk.E),
        }),
    ('aspect', _default_spinbox_prop),
    ('autoseparators', _default_true_false_prop),
    ('background', _default_color_prop),
    ('before', _empty_choice),
    ('bitmap', {
        'input_method': 'choice',
        'values': ('', 'error', 'gray75', 'gray50', 'gray25', 'gray12',
            'hourglass', 'info', 'questhead', 'question', 'warning')
        }),
    ('borderwidth', _default_dimension_prop),
    ('buttonbackground', _default_color_prop),
    ('buttoncursor', _default_cursor_prop),
    ('buttondownrelief', _default_relief_prop),
    ('buttonup', _default_relief_prop),
    ('class_', _default_entry_prop),
    ('closeenough', _default_spinbox_prop),
    ('columnbreak', _default_true_false_prop),
    ('command', _default_entry_prop),
    ('command_id_arg', {
        'input_method': 'choice',
        'values': ('True', 'False'),
        'default': 'False'}),
    ('compound', {
        'input_method': 'choice',
        'values': ('', tk.TOP, tk.BOTTOM,
            tk.LEFT, tk.RIGHT),
        'tk.Radiobutton': {
            'values': ('', tk.NONE, tk.TOP, tk.BOTTOM,
                tk.LEFT, tk.RIGHT)},
        'tk.Menubutton': {
            'values': ('', tk.NONE, tk.TOP, tk.BOTTOM,
                tk.LEFT, tk.RIGHT)},
        'ttk.Label': {
            'values' : ('', tk.BOTTOM, 'image', tk.LEFT, 'none',
                tk.RIGHT, 'text', tk.TOP)},
        }),
    ('confine', _default_true_false_prop),
    ('cursor', _default_cursor_prop),
    ('digits', _default_spinbox_prop),
    ('default', {
        'input_method': 'choice',
        'values': (tk.NORMAL, tk.DISABLED)
        }),
    ('direction', {
        'input_method': 'choice', 'values': None,
        'tk.Menubutton': {
            'values': ('', tk.LEFT, tk.RIGHT, 'above')},
        'ttk.Menubutton': {
            'values': ('', 'above', 'below', 'flush',
                tk.LEFT, tk.RIGHT)},
        }),
    ('disabledbackground', _default_color_prop),
    ('disabledforeground', _default_color_prop),
    ('elementborderwidth', {'input_method': 'spinbox', 'min': -1, 'max': 99,
        'validator': 'number_integer'}),
    ('exportselection', _default_true_false_prop),
    ('font', _default_entry_prop),
    ('foreground', _default_color_prop),
    ('format', _default_entry_prop),
    ('from_', {
            'input_method': 'spinbox',
            'min': -999, 'max': 999, 'default': 0,
            'validator': 'number_float',
            'increment': 0.1
        }),
    ('to', {
            'input_method': 'spinbox',
            'min': -999, 'max': 999, 'default': 100,
            'validator': 'number_float',
            'increment': 0.1
        }),
    ('increment', _default_spinbox_float_prop),
    ('handlepad', _default_dimension_prop),
    ('handlesize', _default_dimension_prop),
    ('hidemargin', _default_true_false_prop),
    ('height', {
        'input_method': 'spinbox', 'min': 0, 'max': 999,
        'validator': 'number_integer',
        'tk.Toplevel': { 'default': 200 },
        'tk.Frame': { 'default': 200 },
        'ttk.Frame': { 'default': 200 },
        'tk.LabelFrame': { 'default': 200 },
        'ttk.Labelframe': { 'default': 200 },
        'tk.PanedWindow': { 'default': 200 },
        'ttk.Panedwindow': { 'default': 200 },
        'ttk.Notebook': { 'default': 200 },
        'tk.Text': { 'default': 10 },
        'pygubu.builder.widgets.dialog': { 'default': 100 },
        }), #FIXME this prop has diferent interpretations
    ('width', {
        'input_method': 'spinbox', 'min': 0, 'max': 999,
        'validator': 'number_integer',
        'tk.Toplevel': { 'default': 200 },
        'tk.Frame': { 'default': 200 },
        'ttk.Frame': { 'default': 200 },
        'tk.LabelFrame': { 'default': 200 },
        'ttk.Labelframe': { 'default': 200 },
        'tk.PanedWindow': { 'default': 200 },
        'ttk.Panedwindow': { 'default': 200 },
        'ttk.Notebook': { 'default': 200 },
        'tk.Text': { 'default': 50 },
        'ttk.Treeview.Column': {'min':5, 'default': 200 },
        'pygubu.builder.widgets.dialog': { 'default': 200 },
        }), #FIXME width is not a dimension for Entry
    ('highlightbackground', _default_color_prop),
    ('highlightcolor', _default_color_prop),
    ('highlightthickness', _default_spinbox_prop),
    ('indicatoron', _default_true_false_prop),
    ('insertbackground', _default_color_prop),
    ('insertborderwidth', _default_dimension_prop),
    ('insertofftime', _default_spinbox_prop),
    ('insertontime', _default_spinbox_prop),
    ('insertwidth', _default_dimension_prop),
    ('image', _default_image_prop),
    ('jump', {'input_method': 'choice', 'values': ('', '0', '1')}),
    ('justify', {
        'input_method': 'choice',
        'values': ('', tk.LEFT, tk.CENTER,
            tk.RIGHT),
        }),
    ('label', _default_entry_prop),
    ('labelanchor', {
        'input_method': 'choice',
        'values': ('', tk.NW, tk.N, tk.NE,
            tk.E + tk.N, tk.E, tk.E + tk.S,
            tk.W + tk.N, tk.W, tk.W + tk.S,
            tk.SW, tk.S, tk.SE)
        }),
    ('labelwidget', _empty_choice),
    ('length', _default_dimension_prop),
    ('listvariable', _default_entry_prop),
    ('maximum', _default_spinbox_prop),
    ('maxundo', {
        'input_method': 'spinbox', 'min':-1, 'max':999, 'default': ''}),
    # maxize: method from Toplevel as property
    ('maxsize', {'input_method': 'sizeentry', 'mode':'whsize'}),
    ('minsize', {'input_method': 'sizeentry',
        'tk.Toplevel': {'mode':'whsize'},
        'pygubu.builder.widgets.dialog': {'mode':'whsize'}
        }),
    ('minwidth', {
        'input_method': 'spinbox', 'min':5, 'max':999, 'default': '20'}),
    ('mode', { 'input_method': 'choice',
        'values': ('', 'determinate', 'indeterminate')}),
    ('offrelief', _default_relief_prop),
    ('offvalue', _default_entry_prop),
    ('onvalue', _default_entry_prop),
    ('opaqueresize', _default_true_false_prop),
    ('orient', {
        'input_method': 'choice',
        'values': (tk.VERTICAL, tk.HORIZONTAL),
        'default': tk.HORIZONTAL
        }),
    ('overrelief', _default_relief_prop),
    ('padding', {'input_method': 'entry', 'validator': 'tkpadding'}),
    ('padx', _default_spinbox_prop),
    ('pady', _default_spinbox_prop),
    ('postcommand', _default_entry_prop),
    ('readonlybackground', _default_color_prop),
    ('relief', _default_relief_prop),
    ('repeatdelay', _default_spinbox_prop),
    ('repeatinterval', _default_spinbox_prop),
    ('resolution', _default_spinbox_float_prop),
    ('scrollregion', _default_entry_prop),
    ('sashpad', _default_dimension_prop),
    ('sashrelief', _default_relief_prop),
    ('sashwidth', _default_dimension_prop),
    ('selectcolor', _default_color_prop),
    ('selectbackground', _default_color_prop),
    ('selectborderwidth', _default_spinbox_prop),
    ('selectforeground', _default_color_prop),
    ('selectimage', _default_image_prop),
    ('selectmode', {
        'input_method': 'choice',
        'values': ('', tk.BROWSE, tk.SINGLE,
            tk.MULTIPLE, tk.EXTENDED),
        'ttk.Treeview': {
            'values': (tk.EXTENDED, tk.BROWSE, tk.NONE),
            'default': tk.EXTENDED,
            }
        }),
    ('show', {
        'input_method': 'choice',
        'tk.Entry': {
            'values': ('', '•'),
            'readonly': False,
            'default': ''},
        'ttk.Entry': {
            'values': ('', '•'),
            'readonly': False,
            'default': ''},
        'ttk.Treeview': {
            'values': ('', 'tree', 'headings'),
            'default': '',
            'readonly': True},
        'pygubu.builder.widgets.editabletreeview': {
            'values': ('', 'tree', 'headings'),
            'default': '',
            'readonly': True},
        }),
    ('showhandle', _default_true_false_prop),
    ('showvalue', _default_true_false_prop),
    ('sliderlength', _default_dimension_prop),
    ('sliderrelief', _default_relief_prop),
    ('spacing1', _default_spinbox_prop),
    ('spacing2', _default_spinbox_prop),
    ('spacing3', _default_spinbox_prop),
    ('state', {
        'input_method': 'choice',
        'values': ('', tk.NORMAL, tk.DISABLED),
        'tk.Entry': {
            'values': ('', tk.NORMAL, tk.DISABLED, 'readonly')},
        'tk.Combobox': {'values': ('', 'readonly')},
        'ttk.Entry': {
            'values': ('', tk.NORMAL, tk.DISABLED, 'readonly')},
        'ttk.Combobox': { 'values': ('', 'readonly')},
        }),
    ('sticky', _sticky_prop),
    ('stretch', {
        'input_method': 'choice',
        'values': ('True', 'False'),
        'default': 'True',
        'readonly': True}),
    ('style', {
        'input_method': 'entry'
        }), #FIXME Howto manage this property?
    ('tabs', _default_entry_prop), #FIXME see tk.Text tab property
    ('takefocus', _default_true_false_prop),
    ('tearoff', { 'input_method': 'choice',
        'values': ('', tk.TRUE, tk.FALSE),
        'tk.Men': {'default': tk.FALSE},
        'tk.Menuitem.Submen': {'default': tk.FALSE}
        }),
    ('tearoffcommand', _default_entry_prop),
    ('text', _default_textentry_prop),
    ('textvariable', _default_tkvariable_prop),
    ('tickinterval', _default_spinbox_float_prop),
    ('title', _default_entry_prop),
    ('troughcolor', _default_color_prop),
    ('undo', _default_true_false_prop),
    ('underline', _default_spinbox_prop),
    ('validate', { 'input_method': 'choice',
        'values': ('none', 'focus', 'focusin', 'focusout', 'key', 'all'),
        'default': 'none'
        }),
    ('validatecommand', _default_entry_prop),
    ('validatecommand_args', { 'input_method': 'entry',
        'validator': 'entry_validate_args' }),
    ('invalidcommand', _default_entry_prop),
    ('invalidcommand_args',  { 'input_method': 'entry',
        'validator': 'entry_validate_args' }),
    ('value', { 'input_method': 'entry', 'validator': 'alphanumeric',
        'ttk.Scale': { 'validator': 'number_float'}
        }),
    ('values', _default_entry_prop), #FIXME This should be treated as a list?
    ('variable',_default_tkvariable_prop),
    ('weight', _default_spinbox_prop),
    ('wrap', { 'input_method': 'choice',
        'values': ('', tk.TRUE, tk.FALSE),
        'tk.Text': {
            'values': (tk.CHAR, tk.WORD, tk.NONE),
            'default': tk.CHAR}
        }),
    ('wraplength', _default_dimension_prop),
    ('xscrollcommand', _default_entry_prop),
    ('xscrollincrement', _default_dimension_prop),
    ('yscrollcommand', _default_entry_prop),
    ('yscrollincrement', _default_dimension_prop),
    #toplevel methods as properties
    ('title', _default_entry_prop),
    ('geometry', _default_entry_prop),
    ('overrideredirect', _default_true_false_prop),
    ('resizable', { 'input_method': 'choice',
        'values': ('', 'both', 'horizontally', 'vertically', 'none'),
        'readonly': True,
        }),
)
PropertiesMap[GROUP_WIDGET] = OrderedDict(__widget)

__grid = (
#grid packing properties
    ('row', {
        'input_method': 'spinbox',
        'min': 0, 'max': 50, 'validator': 'number_integer'}),
    ('column', {
        'input_method': 'spinbox',
        'min': 0, 'max': 50, 'validator': 'number_integer' }),
    ('sticky', _sticky_prop),
    ('rowspan', {
        'input_method': 'spinbox',
        'min': 1, 'max': 50, 'validator': 'number_integer', 'readonly': True}),
    ('columnspan', {
        'input_method': 'spinbox',
        'min': 1, 'max': 50, 'validator': 'number_integer', 'readonly': True}),
    ('padx', {'input_method':'entry', 'validator': 'tkpadding2'}),
    ('pady', {'input_method':'entry', 'validator': 'tkpadding2'}),
    ('ipadx', _default_spinbox_prop),
    ('ipady', _default_spinbox_prop),
    ('propagate', {'input_method': 'choice',
        'values': ('True', 'False'), 'default': 'True'})
)

PropertiesMap[GROUP_LAYOUT_GRID] = OrderedDict(__grid)

__grid_rc = (
    #grid row and column properties (can be applied to each row or column)
    ('minsize', {
        'input_method': 'spinbox',
        'min': 0, 'max':999, 'validator': 'number_integer', 'readonly': True}),
    ('pad', {
        'input_method': 'spinbox',
        'min': 0, 'validator': 'number_integer', 'readonly': True}),
    ('weight', {
        'input_method': 'spinbox',
        'min': 0, 'validator': 'number_integer', 'readonly': True})
)

PropertiesMap[GROUP_LAYOUT_GRID_RC] = OrderedDict(__grid_rc)

__custom = (
    ('class', {
        'input_method': 'entry',
        'readonly': True
        }),
    ('id', {'input_method': 'entry'}),
    ('tree_column', {
        'input_method': 'choice',
        'values': ('True', 'False'),
        'default': 'False',
        'readonly': True
        }),
    ('visible', {
        'input_method': 'choice',
        'values': ('True', 'False'),
        'default': 'True',
        'readonly': True}),
    ('heading_anchor', {
        'input_method': 'choice',
        'values': ('', tk.W, tk.CENTER, tk.E),
        'default': tk.W,
        'readonly': True
        }),
    ('column_anchor', {
        'input_method': 'choice',
        'values': ('', tk.W, tk.CENTER, tk.E),
        'default': tk.W,
        'readonly': True
        }),
    ('scrolltype', {
        'input_method': 'choice',
        'values': ('both', 'vertical', 'horizontal'),
        'default': 'both',
        'readonly': True
        }),
)

PropertiesMap[GROUP_CUSTOM] = OrderedDict(__custom)

def register_custom(name, descr):
    if name not in PropertiesMap[GROUP_CUSTOM]:
        PropertiesMap[GROUP_CUSTOM][name] = descr
    else:
        raise ValueError('Property "{}" already registered'.format(name))


OBJECT_DEFAULT_ATTRS = ('class', 'id')

TRANSLATABLE_PROPERTIES = [
    'label', 'text', 'title',
]



# encoding: UTF-8
#
# Copyright 2012-2013 Alejandro Autalán
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  http://pygubu.web.here

from __future__ import unicode_literals
try:
    import tkinter as tk
    import tkinter.ttk as ttk
    import tkinter.colorchooser
except:
    import Tkinter as tk
    import ttk
    import tkColorChooser
    tk.colorchooser = tkColorChooser

from pygubudesigner.widgets.compoundpropertyeditor import CompoundPropertyEditor
#from compoundpropertyeditor import CompoundPropertyEditor


class ColorEntry(CompoundPropertyEditor):
    count = 0
    ttk_style = None
    btn_bgcolor = ''

    def __init__(self, master=None, **kw):
        CompoundPropertyEditor.__init__(self, master, **kw)
        
        if ColorEntry.ttk_style is None:
            ColorEntry.ttk_style = ttk.Style()
            ColorEntry.btn_bgcolor = ColorEntry.ttk_style.lookup('Toolbutton', 'background')

        ColorEntry.count += 1

        self._entryvar = tk.StringVar()
        self.entry = w = ttk.Entry(self, textvariable=self._entryvar)
        w.grid(sticky='ew')
        w.bind('<FocusOut>', self._on_entry_changed)
        w.bind('<KeyPress-Return>', self._on_entry_changed)
        w.bind('<KeyPress-KP_Enter>', self._on_entry_changed)
        selector_id = ColorEntry.count
        self.stylename = 'ID{0}.ColorSelectorButton.Toolbutton'.format(selector_id)
        self.button = w = ttk.Button(self, text='…', style=self.stylename)
        w.grid(row=0, column=1, padx="5 0")
        w.configure(command=self._on_button_click)
        
        self.columnconfigure(0, weight=1)

    def _on_button_click(self):
        current = self._entryvar.get()
        txtcolor = None
        try:
            _, txtcolor = tk.colorchooser.askcolor(color=current)
        except tk.TclError:
            pass
        self._change_color(txtcolor)


    def _change_color(self, newcolor, gen_event=True):
        if newcolor:
            try:
                rgb = self.winfo_rgb(newcolor)
                ColorEntry.ttk_style.configure(self.stylename, background=newcolor)
            except tk.TclError as e:
                pass
        else:
            ColorEntry.ttk_style.configure(self.stylename,
                background=ColorEntry.btn_bgcolor)
            newcolor = ''
        self._disable_cb()
        self._variable.set(newcolor)
        self._entryvar.set(newcolor)
        self._enable_cb()
        if gen_event:
            self.event_generate('<<ColorEntryChanged>>')

    def _on_entry_changed(self, event=None):
        color = self._entryvar.get()
        self._change_color(color)

    def _on_variable_changed(self, varname, elementname, mode):
        color = self._variable.get()
        self._change_color(color, gen_event=False)

    
    
if __name__ == '__main__':
    root = tk.Tk()
    var = tk.StringVar()
    
    def see_var():
        print(var.get())

    entry = ColorEntry(root, textvariable=var)
    entry.grid()
    entry.configure(textvariable=var)
    var.set('red')
    
    btn = ttk.Button(root, text='Value', command=see_var)
    btn.grid(row=0, column=1)

    root.mainloop()




# encoding: UTF-8
#
# Copyright 2012-2013 Alejandro Autalán
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  http://pygubu.web.here

from __future__ import unicode_literals
from collections import OrderedDict
import xml.etree.ElementTree as ET
try:
    import tkinter as tk
    import tkinter.ttk as ttk
except:
    import Tkinter as tk
    import ttk

import pygubu
from pygubu.stockimage import StockImage, StockImageException
from . import util
import pygubudesigner.widgets.toplevelframe


class Preview(object):

    def __init__(self, id_, canvas, x=0, y=0):
        self.id = 'preview_{0}'.format(id_)
        self.x = x
        self.y = y
        self.w = 10
        self.h = 10
        self.min_w = self.w
        self.min_h = self.h
        self.resizer_h = 10
        self.canvas = canvas
        self.shapes = {}
        self._create_shapes()
        #--------
        self.builder = None
        self.canvas_window= None

    def width(self):
        return self.w

    def height(self):
        return self.h + self.resizer_h

    def _create_shapes(self):
        #Preview box
        c = self.canvas
        x, y, x2, y2 = (-1001, -1000, -1001, -1000)
        s1 = c.create_rectangle(x, y, x2, y2,
            width=2, outline='blue', tags=self.id)
        s2 = c.create_rectangle(x, y, x2, y2, fill='blue', outline='blue',
            tags=(self.id, 'resizer'))
        s3 = c.create_text(x, y, text='widget_id', anchor=tk.NW,
            fill='white', tags=self.id)
        s4 = c.create_window(x, y, anchor=tk.NW, tags=self.id)
        self.shapes = {
            'outline': s1,
            'resizer': s2,
            'text': s3,
            'window': s4
        }
        self.draw()

    def erase(self):
        self.canvas_window.destroy()
        for key, sid in self.shapes.items():
            self.canvas.delete(sid)

    def draw(self):
        c = self.canvas
        x, y, x2, y2 = (self.x, self.y, self.x + self.w, self.y + self.h)
        c.coords(self.shapes['outline'], x, y, x2, y2)
        tbbox = c.bbox(self.shapes['text'])
        tw , th = tbbox[2] - tbbox[0] + 10, tbbox[3] - tbbox[1] + 6
        self.resizer_h = th
        rx2 = self.x + self.w
        ry2 = self.y + self.h + self.resizer_h
        rx = rx2 - tw
        ry = self.y + self.h
        c.coords(self.shapes['resizer'], rx, ry, rx2, ry2)
        tx = rx + 5
        ty = ry + 3
        c.coords(self.shapes['text'], tx, ty)
        c.coords(self.shapes['window'], x, y)

    def move_by(self, dx, dy):
        self.x += dx
        self.y += dy
        self.draw()

    def resize_to(self, w, h):
        self.resize_by(w - self.w, h - self.h)

    def resize_by(self, dw, dh):
        new_w = self.w + dw
        new_h = self.h + dh
        changed = False
        if new_w >= self.min_w:
            self.w = new_w
            changed = True
        if new_h >= self.min_h:
            self.h = new_h
            changed = True
        if changed:
            self.draw()
            self._resize_preview_window()

    def _resize_preview_window(self):
        if self.canvas_window:
            self.canvas_window.configure(width=self.w, height=self.h)

    def update(self, widget_id, xmlnode):
        #delete current preview
        #FIXME maybe do something to update preview without re-creating all ?
        del self.builder
        self.builder = None
        self.canvas.itemconfigure(self.shapes['window'], window='')
        if self.canvas_window:
            self.canvas_window.destroy()

        #Create preview
        canvas_window = ttk.Frame(self.canvas)
        canvas_window.rowconfigure(0, weight=1)
        canvas_window.columnconfigure(0, weight=1)

        self.canvas.itemconfigure(self.shapes['text'], text=widget_id)

        preview_widget = self.create_preview_widget(
            canvas_window, widget_id, xmlnode)

        self.canvas_window = canvas_window
        self.canvas.itemconfigure(self.shapes['window'], window=canvas_window)
        canvas_window.update_idletasks()
        canvas_window.grid_propagate(0)
        self.w = self.min_w = preview_widget.winfo_reqwidth()
        self.h = self.min_h = preview_widget.winfo_reqheight()
        self.resize_to(self.w, self.h)


    def create_preview_widget(self, parent, widget_id, xmlnode):
        self.builder = pygubu.Builder()
        self.builder.add_from_xmlnode(xmlnode)
        widget = self.builder.get_object(widget_id, parent)
        return widget

    def get_widget_by_id(self, widget_id):
        return self.builder.get_object(widget_id)


    def create_toplevel(self, widget_id, xmlnode):
        #Create preview
        builder = pygubu.Builder()
        builder.add_from_xmlnode(xmlnode)
        top = tk.Toplevel(self.canvas)
        top.columnconfigure(0, weight=1)
        top.rowconfigure(0, weight=1)
        builder.get_object(widget_id, top)
        return top


class MenuPreview(Preview):

    def create_preview_widget(self, parent, widget_id, xmlnode):
        self.builder = pygubu.Builder()
        self.builder.add_from_xmlnode(xmlnode)
        menubutton = ttk.Menubutton(parent, text='Menu preview')
        menubutton.grid()
        widget = self.builder.get_object(widget_id, menubutton)
        menubutton.configure(menu=widget)
        return menubutton

    def create_toplevel(self, widget_id, xmlnode):
        #Create preview
        builder = pygubu.Builder()
        builder.add_from_xmlnode(xmlnode)
        top = tk.Toplevel(self.canvas)
        top.columnconfigure(0, weight=1)
        top.rowconfigure(0, weight=1)

        menu = builder.get_object(widget_id, top)
        top['menu'] = menu
        return top

    def resize_by(self, dw, hw):
        return


class ToplevelPreview(Preview):

    def create_preview_widget(self, parent, widget_id, xmlnode):
        xmlnode.set('class', 'pygubudesigner.ToplevelFramePreview')
        layout = ET.Element('layout')
        for n, v in (('row', '0'), ('column', '0'), ('sticky', 'nsew')):
            p = ET.Element('property')
            p.set('name', n)
            p.text = v
            layout.append(p)
        xmlnode.append(layout)
        #print(ET.tostring(xmlnode))
        self.builder = pygubu.Builder()
        self.builder.add_from_xmlnode(xmlnode)
        widget = self.builder.get_object(widget_id, parent)
        return widget

    def create_toplevel(self, widget_id, xmlnode):
        #Create preview
        builder = pygubu.Builder()
        builder.add_from_xmlnode(xmlnode)
        top = builder.get_object(widget_id, self.canvas)
        return top


class DialogPreview(ToplevelPreview):
    def create_toplevel(self, widget_id, xmlnode):
        top = super(DialogPreview, self).create_toplevel(widget_id, xmlnode)
        top.run()
        return top


#    def get_widget_by_id(self, widget_id):
#        return self.canvas_window




class PreviewHelper:
    indicators_tag = ('nw', 'ne', 'sw', 'se')

    def __init__(self, canvas):
        self.canvas = canvas
        self.previews = OrderedDict()
        self.padding = 20
        self.indicators = None
        self._sel_id = None
        self._sel_widget = None
        self.toplevel_previews = []

        self._moving = False
        self._last_event = None
        self._objects_moving = None
        canvas.bind('<Button-1>', self.click_handler)
        canvas.bind('<ButtonRelease-1>', self.release_handler)
        canvas.bind('<Motion>', self.motion_handler)
        canvas.bind('<4>', lambda event : canvas.yview('scroll', -1, 'units'))
        canvas.bind('<5>', lambda event : canvas.yview('scroll', 1, 'units'))
        self._create_indicators()

    def motion_handler(self, event):
        if not self._moving:
            c = event.widget
            x = c.canvasx(event.x)
            y = c.canvasy(event.y)
            if self._over_resizer(x, y):
                c.configure(cursor='fleur')
            else:
                c.configure(cursor='')
        else:
            dx = event.x - self._last_event.x
            dy = event.y - self._last_event.y
            self._last_event = event
            if dx or dy:
                self.resize_preview(dx, dy)

    def click_handler(self, event):
        c = event.widget
        x = c.canvasx(event.x)
        y = c.canvasy(event.y)

        if self._over_resizer(x, y):
            ids = c.find_overlapping(x, y, x, y)
            if ids:
                self._moving = True
                self._objects_moving = ids
                c.configure(cursor='fleur')
                self._last_event = event

    def release_handler(self, event):
        self._objects_moving = None
        self._moving = False

    def _over_resizer(self, x, y):
        "Returns True if mouse is over a resizer"

        over_resizer = False
        c = self.canvas
        ids = c.find_overlapping(x, y, x, y)
        if ids:
            o = ids[0]
            tags = c.gettags(o)
            if 'resizer' in tags:
                over_resizer = True
        return over_resizer

    def resize_preview(self, dw, dh):
        "Resizes preview that is currently dragged"

        #identify preview
        if self._objects_moving:
            id_ = self._objects_moving[0]
            tags = self.canvas.gettags(id_)
            for tag in tags:
                if tag.startswith('preview_'):
                    _, ident = tag.split('preview_')
                    preview = self.previews[ident]
                    preview.resize_by(dw, dh)
                    self.move_previews()
                    break
        self._update_cregion()

    def _update_cregion(self):
        # update canvas scrollregion
        bbox = self.canvas.bbox(tk.ALL)
        padd = 20
        if bbox is not None:
            region = (0, 0, bbox[2] + padd, bbox[3] + padd)
            self.canvas.configure(scrollregion=region)

    def move_previews(self):
        "Move previews after a resize event"

        #calculate new positions
        min_y = self._calc_preview_ypos()
        for idx, (key, p) in enumerate(self.previews.items()):
            new_dy = min_y[idx] - p.y
            self.previews[key].move_by(0, new_dy)
        self._update_cregion()
        self.show_selected(self._sel_id, self._sel_widget)

    def _calc_preview_ypos(self):
        "Calculates the previews positions on canvas"

        y = 10
        min_y = [y]
        for k, p in self.previews.items():
            y += p.height() + self.padding
            min_y.append(y)
        return min_y

    def _get_slot(self):
        "Returns the next coordinates for a preview"

        x = y = 10
        for k, p in self.previews.items():
            y += p.height() + self.padding
        return x, y

    def draw(self, identifier, widget_id, xmlnode, wclass):
        preview_class = Preview
        if wclass == 'tk.Menu':
            preview_class = MenuPreview
        elif wclass == 'tk.Toplevel':
            preview_class = ToplevelPreview
        elif wclass == 'pygubu.builder.widgets.dialog':
            preview_class = DialogPreview
        if identifier not in self.previews:
            x, y = self._get_slot()
            self.previews[identifier] = preview = preview_class(identifier, self.canvas, x, y)
        else:
            preview = self.previews[identifier]
        preview.update(widget_id, xmlnode)
        self.reset_selected(identifier)
        self.move_previews()

    def _create_indicators(self):
        #selected indicators
        self.indicators = []
        anchors = {'nw': tk.SE, 'ne': tk.SW, 'sw': tk.NE, 'se': tk.NW}
        for sufix in self.indicators_tag:
            label = tk.Label(self.canvas,
                    image=StockImage.get('indicator_' + sufix))
            self.indicators.append(label)
            self.canvas.create_window(-10, -10, anchor=anchors[sufix],
                    window=label, tags=sufix)

    def _calculate_indicator_coords(self, tag, widget):
        x = y = 0
        wx = widget.winfo_rootx()
        wy = widget.winfo_rooty()
        ww = widget.winfo_width()
        wh = widget.winfo_height()
        cx = self.canvas.winfo_rootx()
        cy = self.canvas.winfo_rooty()
        if tag == 'nw':
            x = wx - cx
            y = wy - cy
        if tag == 'ne':
            x = (wx - cx) + ww
            y = (wy - cy)
        if tag == 'sw':
            x = (wx - cx)
            y = (wy - cy) + wh
        if tag == 'se':
            x = (wx - cx) + ww
            y = (wy - cy) + wh
        x, y = self.canvas.canvasx(x), self.canvas.canvasy(y)
        return (x, y)

    def show_selected(self, identifier, selected_id=None):
        canvas = self.canvas
        if selected_id is None:
            for indicator in self.indicators_tag:
                canvas.itemconfigure(indicator, state=tk.HIDDEN)
        elif identifier in self.previews:
            for indicator in self.indicators_tag:
                canvas.itemconfigure(indicator, state=tk.NORMAL)
            preview = self.previews[identifier]
            canvas.update_idletasks()
            widget = preview.get_widget_by_id(selected_id)
            for indicatorw in self.indicators:
                try:
                    indicatorw.lift(widget)
                except tk.TclError:
                    pass
            for tag in self.indicators_tag:
                x, y = self._calculate_indicator_coords(tag, widget)
                ox, oy = canvas.coords(tag)
                canvas.move(tag, x - ox, y - oy)
        self._sel_id = identifier
        self._sel_widget = selected_id

    def delete(self, identifier):
        if identifier in self.previews:
            preview = self.previews[identifier]
            preview.erase()
            del self.previews[identifier]
            self.reset_selected(identifier)
            self.move_previews()

    def reset_selected(self, identifier):
        if identifier == self._sel_id:
            self._sel_id = None
            self._sel_widget = None


    def remove_all(self):
        for identifier in self.previews:
            self.delete(identifier)

    def preview_in_toplevel(self, identifier, widget_id, xmlnode):
        preview = self.previews[identifier]
        top = preview.create_toplevel(widget_id, xmlnode)
        self.toplevel_previews.append(top)

    def close_toplevel_previews(self):
        for top in self.toplevel_previews:
            top.destroy()
        self.toplevel_previews = []



from __future__ import unicode_literals
from pygubu.builder.builderobject import *
from pygubu.builder.tkstdwidgets import TKToplevel
from pygubu.widgets.dialog import Dialog


class DialogBO(TKToplevel):
    class_ = Dialog
    
    properties = TKToplevel.properties + ('modal',)
    
    def realize(self, parent):
        BuilderObject.realize(self, parent)
        
    def _set_property(self, target_widget, pname, value):
        if pname == 'modal':
            modal = False
            if value == 'True':
                modal = True
            self.widget.set_modal(modal)
        else:
            super(DialogBO, self)._set_property(self.widget.toplevel, pname, value)

    def get_child_master(self):
        return self.widget.toplevel


register_widget('pygubu.builder.widgets.dialog', DialogBO,
    'Dialog', ('Pygubu Widgets', 'ttk'))


modal_prop = { 'input_method': 'choice',
    'values': ('True', 'False'), 'default':'False', 'readonly': True}

register_property('modal', modal_prop)



from __future__ import unicode_literals
import types

try:
    import tkinter as tk
except:
    import Tkinter as tk

from .builderobject import *

#
# tkinter widgets
#
class TKToplevel(BuilderObject):
    class_ = tk.Toplevel
    container = True
    allow_container_layout = False
    layout_required = False
    allowed_parents = ('root',)
#    allowed_children = ('tk.Frame', 'ttk.Frame')
#    maxchildren = 1
    properties = ('background', 'borderwidth', 'class_', 'cursor', 'height',
        'highlightbackground', 'highlightcolor', 'highlightthickness', 'padx',
        'pady', 'relief', 'takefocus', 'width',
        #Extra properties as methods
        'title', 'geometry', 'overrideredirect', 'minsize', 'maxsize',
        'resizable')
    RESIZABLE = {
        'both': (True, True),
        'horizontally': (True, False),
        'vertically': (False, True),
        'none': (False, False)
        }
        
    def realize(self, parent):
        args = self._get_init_args()
        master = parent.get_child_master()
        if master is None and tk._default_root is None:
            self.widget = tk.Tk()
        else:
            self.widget = self.class_(master, **args)
        return self.widget
        
    def _set_property(self, target_widget, pname, value):
        method_props = ('geometry', 'overrideredirect', 'title')
        if pname in method_props:
            method = getattr(target_widget, pname)
            method(value)
        elif pname == 'resizable' and value:
            target_widget.resizable(*self.RESIZABLE[value])
            if value in ('both', 'horizontally'):
                target_widget.columnconfigure(0, weight=1)
            if value in ('both', 'vertically'):
                target_widget.rowconfigure(0, weight=1)
        elif pname == 'maxsize':
            if '|' in value:
                w, h = value.split('|')
                target_widget.maxsize(w, h)
        elif pname == 'minsize':
            if '|' in value:
                w, h = value.split('|')
                target_widget.minsize(w, h)
        else:
            super(TKToplevel, self)._set_property(target_widget, pname, value)

register_widget('tk.Toplevel', TKToplevel, 'Toplevel', ('Containers', 'tk', 'ttk'))


class TKFrame(BuilderObject):
    class_ = tk.Frame
    container = True
    properties = ['background', 'borderwidth', 'cursor', 'height',
        'highlightbackground', 'highlightcolor', 'highlightthickness',
        'padx', 'pady', 'relief', 'takefocus', 'width']

register_widget('tk.Frame', TKFrame, 'Frame', ('Containers', 'tk'))


class TKLabel(BuilderObject):
    class_ = tk.Label
    container = False
    properties = ['activebackground', 'activeforeground', 'anchor',
        'background', 'bitmap', 'borderwidth', 'compound',
        'cursor', 'disabledforeground', 'font', 'foreground', 'height',
        'highlightbackground', 'highlightcolor', 'highlightthickness',
        'image', 'justify', 'padx', 'pady', 'relief', 'state',
        'takefocus', 'text', 'textvariable', 'underline',
        'width', 'wraplength']

register_widget('tk.Label', TKLabel, 'Label', ('Control & Display', 'tk'))


class TKLabelFrame(BuilderObject):
    class_ = tk.LabelFrame
    container = True
    properties = ['background', 'borderwidth', 'cursor', 'height',
        'highlightbackground', 'highlightcolor', 'highlightthickness',
        'labelanchor', 'labelwidget', 'padx', 'pady', 'relief',
        'text', 'takefocus', 'width']
#TODO: Add helper so the labelwidget can be configured on GUI

register_widget('tk.LabelFrame', TKLabelFrame, 'LabelFrame', ('Containers', 'tk'))


class TKEntry(EntryBaseBO):
    class_ = tk.Entry
    container = False
    properties = ['background', 'borderwidth', 'cursor',
        'disabledbackground', 'disabledforeground', 'exportselection',
        'foreground', 'font', 'highlightbackground', 'highlightcolor',
        'highlightthickness', 'insertbackground', 'insertborderwidth',
        'insertofftime', 'insertontime', 'insertwidth', 'justify',
        'readonlybackground', 'relief', 'selectbackground',
        'selectborderwidth', 'selectforeground', 'show', 'state',
        'takefocus', 'textvariable', 'validate', 'validatecommand',
        'invalidcommand', 'width', 'wraplength', 'xscrollcommand',
        'text', # < text is a custom property
        'validatecommand_args',
        'invalidcommand_args']
    command_properties = ('validatecommand', 'invalidcommand',
         'xscrollcommand')

register_widget('tk.Entry', TKEntry, 'Entry', ('Control & Display', 'tk'))


class TKButton(BuilderObject):
    class_ = tk.Button
    container = False
    properties = ['activebackground', 'activeforeground', 'anchor',
        'borderwidth', 'background', 'bitmap', 'command', 'cursor',
        'default', 'disabledforeground', 'foreground', 'font', 'height',
        'highlightbackground', 'highlightcolor', 'highlightthickness',
        'image', 'justify', 'overrelief', 'padx', 'pady', 'relief',
        'repeatdelay', 'repeatinterval', 'state', 'takefocus', 'text',
        'textvariable', 'underline', 'width', 'wraplength']
    command_properties = ('command',)

register_widget('tk.Button', TKButton, 'Button', ('Control & Display', 'tk'))


class TKCheckbutton(BuilderObject):
    class_ = tk.Checkbutton
    container = False
    properties = ['activebackground', 'activeforeground', 'anchor',
        'background', 'bitmap', 'borderwidth', 'command', 'compound',
        'cursor', 'disabledforeground', 'font', 'foreground', 'height',
        'highlightbackground', 'highlightcolor', 'highlightthickness',
        'image', 'indicatoron', 'justify', 'offrelief', 'offvalue',
        'onvalue', 'overrelief', 'padx', 'pady', 'relief', 'selectcolor',
        'selectimage', 'state', 'takefocus', 'text', 'textvariable',
        'underline', 'variable', 'width', 'wraplength']
    command_properties = ('command',)

register_widget('tk.Checkbutton', TKCheckbutton,
    'Checkbutton', ('Control & Display', 'tk'))


class TKListbox(BuilderObject):
    class_ =  tk.Listbox
    container = False
    properties = ['activestyle', 'background', 'borderwidth', 'cursor',
            'disabledforeground', 'exportselection', 'font',
            'foreground', 'height', 'highlightbackground', 'highlightcolor',
            'highlightthickness', 'listvariable', 'relief',
            'selectbackground', 'selectborderwidth', 'selectforeground',
            'selectmode', 'state', 'takefocus', 'width', 'xscrollcommand',
            'yscrollcommand']
    command_properties = ('xscrollcommand', 'yscrollcommand')

register_widget('tk.Listbox', TKListbox, 'Listbox', ('Control & Display', 'tk'))


class TKText(BuilderObject):
    class_ = tk.Text
    container = False
    properties = ['autoseparators', 'background', 'borderwidth', 'cursor',
            'exportselection', 'font',
            'foreground', 'height', 'highlightbackground', 'highlightcolor',
            'highlightthickness', 'insertbackground', 'insertborderwidth',
            'insertofftime', 'insertontime', 'insertwidth',
            'maxundo', 'padx', 'pady', 'relief', 'selectbackground',
            'selectborderwidth', 'selectforeground', 'spacing1',
            'spacing2', 'spacing3', 'state', 'tabs', 'takefocus',
            'undo', 'width', 'wrap', 'xscrollcommand', 'yscrollcommand',
            'text'] #<- text is a custom property.
    command_properties = ('xscrollcommand', 'yscrollcommand')

    def _set_property(self, target_widget, pname, value):
        if pname == 'text':
            target_widget.insert('0.0', value)
        else:
            super(TKText, self)._set_property(target_widget,pname, value)


register_widget('tk.Text', TKText, 'Text', ('Control & Display', 'tk', 'ttk'))



class TKPanedWindow(PanedWindow):
    class_ = tk.PanedWindow
    allowed_children = ('tk.PanedWindow.Pane',)
    properties = ['background', 'borderwidth', 'cursor', 'handlepad',
        'handlesize', 'height', 'opaqueresize', 'orient', 'relief',
        'sashpad', 'sashrelief', 'sashwidth', 'showhandle', 'width']

register_widget('tk.PanedWindow', TKPanedWindow,
        'PanedWindow', ('Containers', 'tk'))


class TKMenubutton(BuilderObject):
    class_ = tk.Menubutton
    container = False
    properties = ['activebackground', 'activeforeground', 'anchor',
        'background', 'bitmap', 'borderwidth', 'compound', 'cursor',
        'direction', 'disabledforeground', 'foreground', 'font',
        'height', 'highlightbackground', 'highlightcolor',
        'highlightthickness', 'image', 'justify', 'padx',
        'pady', 'relief', 'state', 'takefocus', 'text', 'textvariable',
        'underline', 'width', 'wraplength']
    allowed_children = ('tk.Menu',)
    maxchildren = 1

    def add_child(self, bobject):
        self.widget.configure(menu=bobject.widget)

register_widget('tk.Menubutton', TKMenubutton, 'Menubutton', ('Control & Display', 'tk',))


class TKMessage(BuilderObject):
    class_ = tk.Message
    container = False
    properties = ['aspect', 'background', 'borderwidth', 'cursor',
        'font', 'foreground', 'highlightbackground', 'highlightcolor',
        'highlightthickness', 'justify', 'padx', 'pady', 'relief',
        'takefocus', 'text', 'textvariable', 'width']

register_widget('tk.Message', TKMessage,
        'Message', ('Control & Display', 'tk', 'ttk'))


class TKRadiobutton(BuilderObject):
    class_ = tk.Radiobutton
    container = False
    properties = ['activebackground', 'activeforeground', 'anchor',
        'background', 'bitmap', 'borderwidth', 'command', 'compound',
        'cursor', 'disabledforeground', 'font', 'foreground', 'height',
        'highlightbackground', 'highlightcolor', 'highlightthickness',
        'image', 'indicatoron', 'justify', 'offrelief', 'overrelief',
        'padx', 'pady', 'relief', 'selectcolor', 'selectimage', 'state',
        'takefocus', 'text', 'textvariable',
        'underline', 'variable', 'width', 'wraplength']
    command_properties = ('command',)

register_widget('tk.Radiobutton', TKRadiobutton,
        'Radiobutton', ('Control & Display', 'tk'))


class TKScale(BuilderObject):
    class_ = tk.Scale
    container = False
    properties = ['activebackground', 'background', 'borderwidth', 'command',
        'cursor', 'digits', 'font', 'foreground', 'from_',
        'highlightbackground', 'highlightcolor', 'highlightthickness',
        'label', 'length', 'orient', 'relief', 'repeatdelay', 'repeatinterval',
        'resolution', 'showvalue', 'sliderlength', 'sliderrelief', 'state',
        'takefocus', 'tickinterval', 'to', 'troughcolor', 'variable', 'width']
    command_properties = ('command',)

register_widget('tk.Scale', TKScale, 'Scale', ('Control & Display', 'tk'))


class TKScrollbar(BuilderObject):
    class_ = tk.Scrollbar
    container = False
    properties = ['activebackground', 'activerelief', 'background',
        'borderwidth', 'command', 'cursor', 'elementborderwidth',
        'highlightbackground', 'highlightcolor', 'highlightthickness',
        'jump', 'orient', 'relief', 'repeatdelay', 'repeatinterval',
        'takefocus', 'troughcolor', 'width']
    command_properties = ('command',)

register_widget('tk.Scrollbar', TKScrollbar,
        'Scrollbar', ('Control & Display', 'tk'))


class TKSpinbox(BuilderObject):
    class_ = tk.Spinbox
    container = False
    properties = ['activebackground', 'background', 'borderwidth',
        'buttonbackground', 'buttoncursor', 'buttondownrelief', 'buttonup',
        'command', 'cursor', 'disabledbackground', 'disabledforeground',
        'exportselection', 'font', 'foreground', 'format', 'from_',
        'highlightbackground', 'highlightcolor', 'highlightthickness',
        'increment', 'insertbackground', 'insertborderwidth',
        'insertofftime', 'insertontime', 'insertwidth', 'justify',
        'readonlybackground', 'relief', 'repeatdelay', 'repeatinterval',
        'selectbackground', 'selectborderwidth', 'selectforeground', 'state',
        'takefocus', 'textvariable', 'to', 'values', 'width', 'wrap',
        'xscrollcommand']
    command_properties = ('command', 'xscrollcommand')

    def configure(self):
        #hack to configure 'from_' and 'to' and avoid exception
        if 'from_' in self.properties:
            from_ = float(self.properties['from_'])
            to = float(self.properties.get('to', 0))
            if from_ > to:
                to = from_ + 1
                self.properties['to'] = str(to)
        super(TKSpinbox, self).configure()

register_widget('tk.Spinbox', TKSpinbox, 'Spinbox', ('Control & Display', 'tk'))


class TKMenu(BuilderObject):
    layout_required = False
    allowed_parents = ('root', 'tk.Menubutton', 'ttk.Menubutton')
    allowed_children = ('tk.Menuitem.Submenu', 'tk.Menuitem.Checkbutton',
        'tk.Menuitem.Command', 'tk.Menuitem.Radiobutton',
        'tk.Menuitem.Separator')
    class_ = tk.Menu
    container = True
    allow_container_layout = False
    properties = ['activebackground', 'activeborderwidth', 'activeforeground',
        'background', 'borderwidth', 'cursor', 'disabledforeground',
        'font', 'foreground', 'postcommand', 'relief', 'selectcolor',
        'tearoff', 'tearoffcommand', 'title']
    command_properties = ('postcommand', 'tearoffcommand')
    allow_bindings = False

    def layout(self):
        pass

register_widget('tk.Menu', TKMenu, 'Menu', ('Containers', 'tk', 'ttk'))


class TKCanvas(BuilderObject):
    class_ = tk.Canvas
    container = False
    properties = ['borderwidth', 'background', 'closeenough', 'confine',
        'cursor', 'height', 'highlightbackground', 'highlightcolor',
        'highlightthickness', 'relief', 'scrollregion', 'selectbackground',
        'selectborderwidth', 'selectforeground', 'takefocus', 'width',
        'xscrollincrement', 'xscrollcommand', 'yscrollincrement',
        'yscrollcommand']
    command_properties = ('xscrollcommand', 'yscrollcommand')

register_widget('tk.Canvas', TKCanvas,
        'Canvas', ('Control & Display', 'tk', 'ttk'))


#
# Helpers for Standard tk widgets
#

class TKMenuitem(BuilderObject):
    class_ = None
    container = False
    itemtype = None
    layout_required = False
    properties = ['accelerator', 'activebackground', 'activeforeground',
        'background', 'bitmap', 'columnbreak', 'command', 'compound',
        'font', 'foreground', 'hidemargin', 'image', 'label', 'state',
        'underline',
        'command_id_arg', #<< custom property !!
        ]
    command_properties = ('command',)
    allow_bindings = False

    def realize(self, parent):
        self.widget = master = parent.widget
        itemproperties = dict(self.properties)
        pname = 'command_id_arg'
        if pname in itemproperties:
            itemproperties.pop(pname)
        pname = 'variable'
        if pname in itemproperties:
            varname = itemproperties[pname]
            itemproperties[pname] = self.builder.create_variable(pname)
        for imageprop in ('image', 'selectimage'):
            if imageprop in itemproperties:
                name = itemproperties[imageprop]
                itemproperties[imageprop] = self.builder.get_image(name)
        master.add(self.itemtype, **itemproperties)
        self.__index = master.index(tk.END)
        return self.widget

    def configure(self):
        pass

    def layout(self):
        pass


    def _create_callback(self, cpname, callback):
        command = callback
        include_id = self.properties.get('command_id_arg', 'False')
        if include_id != 'False':
            def item_callback(item_id=self.objectid):
                callback(item_id)
            command = item_callback
        return command

    def _connect_command(self, cpname, callback):
        self.widget.entryconfigure(self.__index, command=callback)


class TKMenuitemSubmenu(TKMenu):
    allowed_parents = ('tk.Menu', 'tk.Menuitem.Submenu')
    allowed_children = ('tk.Menuitem.Submenu', 'tk.Menuitem.Checkbutton',
        'tk.Menuitem.Command', 'tk.Menuitem.Radiobutton',
        'tk.Menuitem.Separator')
    properties = list(set(TKMenu.properties + TKMenuitem.properties))

    def realize(self, parent):
        master = parent.widget
        menu_properties = dict((k, v) for k, v in self.properties.items()
            if k in TKMenu.properties)

        item_properties = dict((k, v) for k, v in self.properties.items()
            if k in TKMenuitem.properties and k != 'command_id_arg')

        self.widget = submenu = TKMenu.class_(master, **menu_properties)
        item_properties['menu'] = submenu
        master.add(tk.CASCADE, **item_properties)
        return self.widget


    def configure(self):
        pass

    def layout(self):
        pass

register_widget('tk.Menuitem.Submenu', TKMenuitemSubmenu,
        'Menuitem.Submenu', ('Pygubu Helpers', 'tk', 'ttk'))


class TKMenuitemCommand(TKMenuitem):
    allowed_parents = ('tk.Menu', 'tk.Menuitem.Submenu')
    itemtype = tk.COMMAND

register_widget('tk.Menuitem.Command', TKMenuitemCommand,
    'Menuitem.Command', ('Pygubu Helpers', 'tk', 'ttk'))


class TKMenuitemCheckbutton(TKMenuitem):
    allowed_parents = ('tk.Menu', 'tk.Menuitem.Submenu')
    itemtype = tk.CHECKBUTTON
    properties = TKMenuitem.properties + \
            ['indicatoron', 'selectcolor', 'selectimage', 'variable']

register_widget('tk.Menuitem.Checkbutton', TKMenuitemCheckbutton,
    'Menuitem.Checkbutton', ('Pygubu Helpers', 'tk', 'ttk'))


class TKMenuitemRadiobutton(TKMenuitem):
    allowed_parents = ('tk.Menu', 'tk.Menuitem.Submenu')
    itemtype = tk.RADIOBUTTON
    properties = TKMenuitemCheckbutton.properties + \
            ['onvalue', 'offvalue', 'value']

register_widget('tk.Menuitem.Radiobutton', TKMenuitemRadiobutton,
    'Menuitem.Radiobutton', ('Pygubu Helpers', 'tk', 'ttk'))


class TKMenuitemSeparator(TKMenuitem):
    allowed_parents = ('tk.Menu', 'tk.Menuitem.Submenu')
    itemtype = tk.SEPARATOR
    properties = []
    command_properties = tuple()

register_widget('tk.Menuitem.Separator', TKMenuitemSeparator,
        'Menuitem.Separator', ('Pygubu Helpers','tk', 'ttk'))


class TKPanedWindowPane(PanedWindowPane):
    class_ = None
    container = True
    allowed_parents = ('tk.PanedWindow',)
    maxchildren = 1
    properties = ['height', 'minsize', 'padx', 'pady', 'sticky']

register_widget('tk.PanedWindow.Pane', TKPanedWindowPane,
    'PanedWindow.Pane', ('Pygubu Helpers', 'tk'))



from __future__ import unicode_literals
import types
from collections import OrderedDict

try:
    import tkinter as tk
    import tkinter.ttk as ttk
except:
    import Tkinter as tk
    import ttk

from .builderobject import *


#
# ttk widgets
#

class TTKFrame(BuilderObject):
    class_ = ttk.Frame
    container = True
    properties = ['class_', 'cursor', 'height', 'padding',
            'relief', 'style', 'takefocus', 'width']
    ro_properties = ('class_',)

register_widget('ttk.Frame', TTKFrame, 'Frame', ('Containers', 'ttk'))


class TTKLabel(BuilderObject):
    class_ = ttk.Label
    container = False
    properties = ['anchor', 'background', 'borderwidth',
            'class_', 'compound', 'cursor', 'font', 'foreground',
            'image', 'justify', 'padding', 'relief',
            'style', 'takefocus', 'text', 'textvariable', 'underline',
            'width', 'wraplength']
    ro_properties = ('class_',)

register_widget('ttk.Label', TTKLabel, 'Label', ('Control & Display', 'ttk'))


class TTKButton(BuilderObject):
    class_= ttk.Button
    container= False
    properties = ['class_', 'command', 'compound', 'cursor',
            'image', 'style', 'takefocus', 'text', 'textvariable',
            'underline']
    ro_properties = ('class_',)
    command_properties = ('command',)

register_widget('ttk.Button', TTKButton, 'Button', ('Control & Display', 'ttk'))


class TTKCheckbutton(BuilderObject):
    class_ = ttk.Checkbutton
    container = False
    properties = ['class_', 'command', 'compound', 'cursor',
            'image', 'style', 'takefocus', 'text', 'textvariable',
            'underline', 'variable', 'offvalue', 'onvalue', 'width']
    ro_properties = ('class_',)
    command_properties = ('command',)

register_widget('ttk.Checkbutton', TTKCheckbutton,
        'Checkbutton', ('Control & Display', 'ttk'))


class TTKRadiobutton(BuilderObject):
    class_ = ttk.Radiobutton
    container = False
    properties = ['class_', 'command', 'compound', 'cursor',
            'image', 'style', 'takefocus', 'text', 'textvariable',
            'underline', 'value', 'variable', 'width']
    ro_properties = ('class_',)
    command_properties = ('command',)

register_widget('ttk.Radiobutton', TTKRadiobutton,
        'Radiobutton', ('Control & Display', 'ttk'))


class TTKCombobox(BuilderObject):
    class_ = ttk.Combobox
    container = False
    properties = ['class_', 'cursor', 'exportselection',
            'height', 'justify', 'postcommand', 'style', 'takefocus',
            'textvariable', 'validate', 'validatecommand', 'invalidcommand',
            'values', 'width', 'xscrollcommand', 'state',
            'validatecommand_args',
            'invalidcommand_args']
    ro_properties = ('class_',)
    command_properties = ('postcommand', 'validatecommand',
        'invalidcommand', 'xscrollcommand')

    def _set_property(self, target_widget, pname, value):
        if pname in ('validatecommand_args', 'invalidcommand_args'):
            pass
        else:
            super(TTKCombobox, self)._set_property(target_widget, pname, value)

    def _create_callback(self, cpname, command):
        callback = command
        if cpname in ('validatecommand', 'invalidcommand'):
            args = self.properties.get(cpname + '_args', '')
            if args:
                args = args.split(' ')
                callback = (self.widget.register(command),) + tuple(args)
            else:
                callback = self.widget.register(command)
        return callback

register_widget('ttk.Combobox', TTKCombobox,
        'Combobox', ('Control & Display', 'ttk'))


class TTKScrollbar(BuilderObject):
    class_ = ttk.Scrollbar
    container = False
    properties = ['class_', 'command', 'cursor', 'orient',
        'style', 'takefocus']
    ro_properties = ('class_',)
    command_properties = ('command',)

register_widget('ttk.Scrollbar', TTKScrollbar,
     'Scrollbar', ('Control & Display', 'ttk'))


class TTKSizegrip(BuilderObject):
    class_ = ttk.Sizegrip
    container = False
    properties = ['class_', 'style']
    ro_properties = ('class_',)

register_widget('ttk.Sizegrip', TTKSizegrip,
        'Sizegrip', ('Control & Display', 'ttk'))


class TTKEntry(EntryBaseBO):
    class_ = ttk.Entry
    container = False
    properties = ['class_', 'cursor', 'exportselection', 'font',
            'invalidcommand', 'justify', 'show', 'state', 'style', 'takefocus',
            'textvariable', 'validate', 'validatecommand',
            #'values',  << Commented, only useful on Combobox widget ?
            'width', 'xscrollcommand',
            'text', # < text is a custom property
            'validatecommand_args',
            'invalidcommand_args']
    ro_properties = ('class_',)
    command_properties = ('validatecommand', 'invalidcommand',
        'xscrollcommand')


register_widget('ttk.Entry', TTKEntry, 'Entry', ('Control & Display', 'ttk'))


class TTKProgressbar(BuilderObject):
    class_ = ttk.Progressbar
    container = False
    properties = ['class_', 'cursor', 'length', 'maximum', 'mode',
            'orient', 'style', 'takefocus', 'variable']
    ro_properties = ('class_',)

register_widget('ttk.Progressbar', TTKProgressbar,
        'Progressbar', ('Control & Display', 'ttk'))


class TTKScale(BuilderObject):
    class_ = ttk.Scale
    container = False
    properties = ['class_', 'command', 'cursor', 'from_', 'length',
            'orient', 'style', 'takefocus', 'to', 'variable', 'value']
    ro_properties = ('class_',)
    command_properties = ('command',)

register_widget('ttk.Scale', TTKScale, 'Scale', ('Control & Display',  'ttk'))


class TTKSeparator(BuilderObject):
    class_ = ttk.Separator
    container = False
    properties = ['class_', 'orient', 'style']
    ro_properties = ('class_',)

register_widget('ttk.Separator', TTKSeparator,
    'Separator', ('Control & Display',  'ttk'))


class TTKLabelframe(BuilderObject):
    class_ = ttk.Labelframe
    container = True
    properties = ['borderwidth', 'class_', 'cursor', 'height',
            'labelanchor', 'labelwidget', 'padding',
            'relief', 'style', 'takefocus', 'text', 'underline', 'width']
    ro_properties = ('class_',)

#TODO: Add helper so the labelwidget can be configured on GUI
register_widget('ttk.Labelframe', TTKLabelframe,
        'Labelframe', ('Containers', 'ttk'))


class TTKPanedwindow(PanedWindow):
    class_ = ttk.Panedwindow
    allowed_children = ('ttk.Panedwindow.Pane',)
    properties = ['class_', 'cursor', 'height', 'orient',
            'style', 'takefocus', 'width']
    ro_properties = ('class_','orient')

register_widget('ttk.Panedwindow', TTKPanedwindow,
        'Panedwindow', ('Containers', 'ttk'))


class TTKNotebook(BuilderObject):
    class_ = ttk.Notebook
    container = True
    allow_container_layout = False
    allowed_children = ('ttk.Notebook.Tab',)
    properties = ['class_', 'cursor', 'height',
            'padding', 'style', 'takefocus', 'width']
    ro_properties = ('class_',)

register_widget('ttk.Notebook', TTKNotebook,
        'Notebook', ('Containers', 'ttk'))


class TTKMenubuttonBO(BuilderObject):
    class_ = ttk.Menubutton
    container = False
    properties = ['class_', 'compound', 'cursor', 'direction',
            'image', 'style', 'takefocus', 'text', 'textvariable',
            'underline', 'width']
    allowed_children = ('tk.Menu',)
    maxchildren = 1
    ro_properties = ('class_',)

    def add_child(self, bobject):
        self.widget.configure(menu=bobject.widget)

register_widget('ttk.Menubutton', TTKMenubuttonBO,
        'Menubutton', ('Control & Display', 'ttk',))


class TTKTreeviewBO(BuilderObject):
    class_ = ttk.Treeview
    container = False
    allowed_children = ('ttk.Treeview.Column',)
    properties = ['class_', 'cursor', 'height', 'padding', 'selectmode',
        'show', 'style', 'takefocus']
    ro_properties = ('class_',)

    def __init__(self, builder, wdescr):
        super(TTKTreeviewBO, self).__init__(builder, wdescr)
        self._columns = None
        self._headings = None
        self._dcolumns = None

    def configure(self):
        super(TTKTreeviewBO,self).configure()
        self.__configure_columns()

    def __configure_columns(self):
        if self._columns:
            columns = list(self._columns.keys())
            if '#0' in columns:
                columns.remove('#0')
            displaycolumns = self._dcolumns
            self.widget.configure(columns=columns,
                                    displaycolumns=displaycolumns)
            for col in self._columns:
                self.widget.column(col, **self._columns[col])
        if self._headings:
            for col in self._headings:
                self.widget.heading(col, **self._headings[col])

    def set_column(self, col_id, attrs, visible=True):
        if self._columns is None:
            self._columns = OrderedDict()
            self._dcolumns = list()
        self._columns[col_id] = attrs
        if visible and col_id != '#0':
            self._dcolumns.append(col_id)

    def set_heading(self, col_id, attrs):
        if self._headings is None:
            self._headings = OrderedDict()
        self._headings[col_id] = attrs


register_widget('ttk.Treeview', TTKTreeviewBO,
        'Treeview', ('Control & Display', 'ttk'))


#
# Helpers for Standard ttk widgets
#

class TTKPanedwindowPane(PanedWindowPane):
    class_ = None
    container = True
    allowed_parents = ('ttk.Panedwindow',)
    maxchildren = 1
    properties = ['weight']

register_widget('ttk.Panedwindow.Pane', TTKPanedwindowPane,
        'Panedwindow.Pane', ('Pygubu Helpers', 'ttk'))


class TTKNotebookTab(BuilderObject):
    class_ = None
    container = True
    layout_required = False
    allow_bindings = False
    allowed_parents = ('ttk.Notebook',)
    maxchildren = 1
    properties = ['compound', 'padding', 'sticky',
        'image', 'text', 'underline']

    def realize(self, parent):
        self.widget = parent.widget
        return self.widget

    def configure(self):
        pass

    def layout(self):
        pass

    def add_child(self, bobject):
        self.widget.add(bobject.widget, **self.properties)

register_widget('ttk.Notebook.Tab', TTKNotebookTab,
        'Notebook.Tab', ('Pygubu Helpers', 'ttk'))


class TTKTreeviewColBO(BuilderObject):
    class_ = None
    container = False
    layout_required = False
    allow_bindings = False
    allowed_parents = ('ttk.Treeview',)
    properties = [
        'tree_column', 'visible', 'text', 'image', 'command', 'heading_anchor',
        'column_anchor', 'minwidth', 'stretch', 'width' ]
    command_properties = ('command',)


    def realize(self, parent):
        self.widget = master = parent.widget

        col_props = dict(self.properties) #copy properties

        tree_column = col_props.pop('tree_column', 'False')
        tree_column = True if tree_column == 'True' else False
        column_id = '#0' if tree_column else self.objectid
        is_visible = True if col_props.pop('visible') == 'True' else False

        #configure heading properties
        command= col_props.pop('command', '')
        hprops = {
            'anchor': col_props.pop('heading_anchor', tk.W),
            'image': col_props.pop('image', ''),
            'text': col_props.pop('text', '')
        }
        parent.set_heading(column_id, hprops)

        #configure column properties
        cprops = {
            'anchor': col_props.pop('column_anchor', ''),
            'stretch': col_props.pop('stretch', '1'),
            'width': col_props.pop('width', '200'),
            'minwidth': col_props.pop('minwidth', '20')
        }
        parent.set_column(column_id, cprops, is_visible)
        return self.widget


    def configure(self):
        pass

    def layout(self):
        pass

    def _connect_command(self, cpname, callback):
        tree_column = self.properties.get('tree_column', 'False')
        tree_column = True if tree_column == 'True' else False
        column_id = '#0' if tree_column else self.objectid
        self.widget.heading(column_id, command=callback)


register_widget('ttk.Treeview.Column', TTKTreeviewColBO,
        'Treeview.Column', ('Pygubu Helpers', 'ttk'))


class TTKSpinboxBO(EntryBaseBO):
    class_ = None
    container = False
    properties = ['class_', 'cursor',
            'from_', 'to', 'increment',
            'values',  #<< Commented, only useful on Combobox widget ?
            'wrap', 'format', 'command',
            'exportselection', 'font',
            'invalidcommand', 'justify', 'show', 'state', 'style', 'takefocus',
            'textvariable', 'validate', 'validatecommand',
            'width', 'xscrollcommand',
            'text', # < text is a custom property
            'validatecommand_args', 'invalidcommand_args']
    ro_properties = ('class_',)
    command_properties = ('validatecommand', 'invalidcommand',
        'xscrollcommand', 'command')


if tk.TkVersion >= 8.6:
    if not hasattr(ttk, 'Spinbox'):
        from pygubu.widgets.ttkspinbox import Spinbox
        ttk.Spinbox = Spinbox

    TTKSpinboxBO.class_ = ttk.Spinbox
    
    register_widget('ttk.Spinbox', TTKSpinboxBO, 'Spinbox', ('Control & Display', 'ttk'))
